import functools
import json
import tempfile
import hashlib
import pathlib
import warnings

warnings.filterwarnings(
    "ignore", message="Using or importing the ABCs", category=DeprecationWarning
)

import attr
import dataset
import sqlalchemy.exc
import sqlite3

from . import core
from . import serializers
from . import autoserial
from . import utils


class NoCache(Exception):
    pass


@attr.s
class MemoryCache:

    key_serializer = attr.ib(default=serializers.json_serializer)
    value_serializer = attr.ib(default=None)
    _data = attr.ib(factory=dict)

    def cache_call(self, function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            call = core.FunctionCall.from_args(function, args, kwargs)
            key = json.dumps(call, default=self.key_serializer, sort_keys=True)

            if key in self._data:
                return self._data[key]

            result = function(*args, **kwargs)
            self._data[key] = result
            return result

        return wrapper


@attr.s
class DirectoryCache:

    key_serializer = attr.ib(serializers.json_serializer)
    value_serializer = attr.ib(default=autoserial.auto_serialize)
    value_deserializer = attr.ib(default=autoserial.auto_deserialize)
    directory = attr.ib(
        factory=lambda: pathlib.Path(tempfile.TemporaryDirectory().name)
    )
    codec = attr.ib(factory=autoserial.AutoCodec)

    # @utils.reify
    @property
    def index(self):
        try:
            return dataset.connect(
                "sqlite:///" + str((self.directory / "index.sqlite").resolve())
            )["file"]
        except (sqlalchemy.exc.OperationalError, sqlite3.OperationalError):
            raise FileNotFoundError

    def cache_call(self, function):
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            call = core.FunctionCall.from_args(function, args, kwargs)
            hashed_key = self.key_serializer(call)
            path = pathlib.Path(self.directory) / hashed_key
            try:
                result = self._deserialize_value(path)
            except NoCache:
                result = function(*args, **kwargs)
                self._serialize_value(result, path)
            return result

        return wrapper

    def _deserialize_value(self, path):
        if not path.exists():
            raise NoCache

        row = self.index.find_one(hash=path.name)
        if row is None:
            raise NoCache
        codec_name = row["codec_name"]

        codec = self.codec._registry[codec_name]

        with open(path, "rb" if codec.binary else "rt") as f:
            return codec.load(f)

    def _serialize_value(self, result, path):
        codec = self.codec._registry[type(result).__name__]

        path.parent.mkdir(exist_ok=True)
        self.index.insert(dict(hash=path.name, codec_name=type(result).__name__))
        with open(path, "wb" if codec.binary else "wt") as f:
            codec.dump(result, f)


@attr.s
class NullCache:
    def cache_call(self, function):
        return function



import json

import attr
import multipledispatch

from . import utils

auto_serialize = utils.ExtensibleDispatcher()
auto_deserialize = utils.ExtensibleDispatcher()


@attr.s
class ValueCodec:
    type = attr.ib()
    name = attr.ib(default=None)
    dump = attr.ib(default=lambda x: x)
    load = attr.ib(default=lambda x: x)
    binary = attr.ib(default=False)


@attr.s
class AutoCodec:
    _registry = attr.ib(factory=dict)

    def serializer(self, type):
        def wrapper(func):
            try:
                value_codec = self._registry[type.__name__]
            except KeyError:
                self._registry[type.__name__] = ValueCodec(
                    type, type.__name__, dump=func
                )
            else:
                self._registry[type.__name__] = attr.evolve(value_codec, dump=func)
            return func

        return wrapper

    def deserializer(self, type):
        def wrapper(func):
            try:
                value_codec = self._registry[type.__name__]
            except KeyError:
                self._registry[type.__name__] = ValueCodec(
                    type, type.__name__, load=func
                )
            else:
                self._registry[type.__name__] = attr.evolve(value_codec, load=func)
            return func

        return wrapper

    def register(self, value_codec):
        self._registry[value_codec.type.__name__] = value_codec



import json

import pytest

from cashmere import backends
from cashmere import serializers
from cashmere import autoserial


def test_json_serialize_list():
    codec = autoserial.AutoCodec()
    codec.register(autoserial.ValueCodec(list, dump=json.dump, load=json.load))
    cache = backends.DirectoryCache(codec=codec)

    lst = [1, 2, 3]

    @cache.cache_call
    def f():
        return lst

    assert f() == lst

    assert f() == lst


def test_json_serialize_fail():
    codec = autoserial.AutoCodec()
    cache = backends.DirectoryCache(codec=codec)

    lst = [1, 2, 3]

    @cache.cache_call
    def f():
        return lst

    with pytest.raises(KeyError):
        assert f() == lst

    with pytest.raises(KeyError):
        assert f() == lst



import collections
import json
import functools
import inspect
import types
import warnings
import enum
import pathlib
import hashlib

import attr
import multipledispatch

from . import core


class NotSerializableWarning(Warning):
    pass


class ErrorAction(enum.Enum):
    RAISE = enum.auto()
    WARN = enum.auto()
    IGNORE = enum.auto()


@attr.s
class Serializable:
    name = attr.ib()
    type = attr.ib()
    value = attr.ib(default=core.UNKNOWN)
    module = attr.ib(default=None)


to_json_serializable = multipledispatch.Dispatcher("registry")


@to_json_serializable.register(object)
def _(obj):
    try:
        return attr.asdict(obj)
    except attr.exceptions.NotAnAttrsClassError:
        return obj


@to_json_serializable.register(core.InputCollection)
def _(obj):
    return collections.OrderedDict(
        [(k, to_json_serializable(v)) for k, v in attr.asdict(obj).items()]
    )


@to_json_serializable.register(types.ModuleType)
def _(obj):
    try:
        string = inspect.getsource(obj)
        type_name = "PythonModule"
    except TypeError:
        string = obj.__name__
        type_name = "BuiltinModule"

    return Serializable(obj.__name__, type_name, string)


@to_json_serializable.register(types.BuiltinFunctionType)
def _(obj):
    return Serializable(
        obj.__name__, type(obj).__name__, value=core.NO_VALUE, module=obj.__module__
    )


@to_json_serializable.register((types.FunctionType, type))
def _(obj):
    recursed = {
        name: to_json_serializable(value)
        for name, value in core.get_closure(obj).items()
    }
    value = {"source": inspect.getsource(obj), "recursed": recursed}
    return Serializable(obj.__name__, type(obj).__name__, value, module=obj.__module__)


@to_json_serializable.register(inspect.BoundArguments)
def _(obj):
    obj.apply_defaults()
    return Serializable("signature", type(obj).__name__, obj.arguments)


@to_json_serializable.register(pathlib.Path)
def _(obj):
    return repr(obj)


@attr.s
class JSONSerializer:
    dispatcher = attr.ib(factory=lambda: multipledispatch.Dispatcher("registry"))
    on_error = attr.ib(default=ErrorAction.RAISE)

    def dispatch(self, type):
        def wrap(function):
            self.dispatcher.register(type)
            return function

        return wrap

    def __call__(self, obj):
        raw_key = json.dumps(obj, default=self._dumps, sort_keys=True)
        hashed_key = hashlib.sha1(raw_key.encode("utf8")).hexdigest()
        return hashed_key

    def _dumps(self, obj):
        try:
            return self.dispatcher(obj)
        except NotImplementedError:
            if self.on_error == ErrorAction.RAISE:
                raise NotImplementedError(obj)
            if self.on_error == ErrorAction.WARN:
                warnings.warn(f"Not serializable: {obj}", NotSerializableWarning)
                return repr(obj)
            if self.on_error == ErrorAction.IGNORE:
                return repr(obj)
            raise

    def copy(self):
        dispatcher = type(self.dispatcher)(
            name=self.dispatcher.name + "_copy", doc=self.dispatcher.doc
        )
        dispatcher.funcs = self.dispatcher.funcs.copy()
        return attr.evolve(self, dispatcher=dispatcher)

    def register(self, *args, **kwargs):
        return self.dispatcher.register(*args, **kwargs)


json_serializer = JSONSerializer(dispatcher=to_json_serializable)



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mquickcache` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``quickcache.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``quickcache.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import click


@click.command()
@click.argument('names', nargs=-1)
def main(names):
    click.echo(repr(names))



from cashmere import core
from cashmere import backends


def test_argument_cache(tmp_path):

    filepath = tmp_path / "demo"
    cache = backends.MemoryCache()

    @cache.cache_call
    def write_into_file(x):
        with open(filepath, "a") as f:
            f.write(x)

    message = "Hello\n"

    # Should write into the file on first invocation.
    write_into_file(message)
    assert filepath.read_text() == message

    # Nothing has changed, so this call should be skipped.
    write_into_file(message)
    assert filepath.read_text() == message

    # Argument has changed, so should write into file.
    second_message = "World\n"
    write_into_file(second_message)
    assert filepath.read_text() == message + second_message


def test_closure_cache(tmp_path):

    filepath = tmp_path / "demo"
    cache = backends.MemoryCache()

    @cache.cache_call
    def write_into_file():
        with open(filepath, "a") as f:
            f.write(message)

    message = "Hello\n"

    # Should write into the file on first invocation.
    write_into_file()
    assert filepath.read_text() == message

    # Nothing has changed, so this call should be skipped.
    write_into_file()
    assert filepath.read_text() == message

    # Closure has changed, so should write into file.
    message = "World\n"
    write_into_file()
    assert filepath.read_text() == "Hello\nWorld\n"



import ast
import collections.abc
import itertools
import collections
import functools
import inspect
import tempfile
import types
import sys
import json
import textwrap

import attr


UNKNOWN = "____UNKNOWN____"
NO_VALUE = None


@attr.s
class InputCollection:

    closure = attr.ib()
    args = attr.ib()


@attr.s
class FunctionCall:
    function = attr.ib()
    inputs = attr.ib()

    @classmethod
    def from_args(cls, function, args, kwargs):
        inputs = _get_inputs(function, *args, **kwargs)

        return cls(function, inputs)


def _extract_referenced_names(function):
    source = textwrap.dedent(inspect.getsource(function))
    tree = ast.parse(source)
    names = {node.id for node in ast.walk(tree) if isinstance(node, ast.Name)}
    parameters = set(inspect.signature(function).parameters)
    return names - parameters


def get_closure(function):
    closure = {}

    for d in reversed(inspect.getclosurevars(function)):

        if isinstance(d, collections.abc.Mapping):
            closure.update(d)

    locally_used_external_names = _extract_referenced_names(function)
    name_to_value = {
        name: closure.get(name, UNKNOWN) for name in locally_used_external_names
    }

    return name_to_value


def _get_inputs(function, *args, **kwargs):

    closure = get_closure(function)
    args = inspect.signature(function).bind(*args, **kwargs)
    args.apply_defaults()

    return InputCollection(closure, args)
